/***************************************************************************
 *   copyright            : (C) 2000 by Marko Grnroos                     *
 *   email                : magi@iki.fi                                    *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#ifndef __AUDITING_H__
#define __AUDITING_H__

#include <magic/Math.h>
#include <magic/Matrix.h>
#include <magic/map.h>
#include <magic/Class.h>
#include <magic/pararr.h>
#include <freenet.h>

class PatternSet;		// In patternset.h
class MatrixEqualizer;	// In equalization.h
class TrainingObserver;	// In trainer.h


/////////////////////////////////////////////////////////////////////////////////////////////////////
//   _             | o     o             -----                 ----                    |           //
//  / \            |    |      _           |    ___   ____  |  |   )  ___   ____       |  |   ____ //
// /   \ |   |  ---| | -+- | |/ \   ___    |   /   ) (     -+- |---  /   ) (     |   | | -+- (     //
// |---| |   | (   | |  |  | |   | (   \   |   |---   \__   |  | \   |---   \__  |   | |  |   \__  //
// |   |  \__!  ---| |   \ | |   |  ---/   |    \__  ____)   \ |  \   \__  ____)  \__! |   \ ____) //
//                                  __/                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////

/** Holder for prediction test results. */
class AuditingTestResults : public Object {
  public:
			AuditingTestResults () {}

	void	make		(int variables);
	void	addValue	(int variable, double value);
	void	calculate	();
	
	double	averageError;	// For all variables
	double	MSE;			// For all variables
	double	RMSE;			// For all variables
	Vector	averageErrors;	// For each variable
	Vector	MSEs;			// For each variable
	Vector	RMSEs;			// For each variable

	void	operator>>	(OStream& out) const;

  private:
	int		mValues; /** Number of values. */
	
	operator = (const AuditingTestResults& o) {FORBIDDEN}
	AuditingTestResults (const AuditingTestResults& o) {FORBIDDEN}
};



/////////////////////////////////////////////////////////////////////////////////////
//   _             | o     o              ----                                     //
//  / \            |    |      _         (      |       ___   |   ___              //
// /   \ |   |  ---| | -+- | |/ \   ___   ---  -+- |/\  ___| -+- /   )  ___  \   | //
// |---| |   | (   | |  |  | |   | (   \     )  |  |   (   |  |  |---  (   \  \  | //
// |   |  \__!  ---| |   \ | |   |  ---/ ___/    \ |    \__|   \  \__   ---/   \_/ //
//                                  __/                                 __/   \_/  //
/////////////////////////////////////////////////////////////////////////////////////

/** Abstract baseclass for different monthly prediction methods.
 *
 *  Design Patterns: Strategy.
 **/
class AuditingStrategy : public Object {
	decl_dynamic (AuditingStrategy);
  public:
						AuditingStrategy	() {FORBIDDEN}
						AuditingStrategy	(const String& name) : mName (name) {}

	/** Initialization. */
	virtual void		make				(const StringMap& params);

	/** Trains the learning method with the given data.
	 *
	 *  Must be implemented by auditing strategies.
	 *
	 *  @param startmonth Zero-based month number for the first sample
	 *  in the data. The month number can be in range 0-11, but only
	 *  the last two digits are interpreted as months, so it can have
	 *  values like 199905 or 9905.
	 **/
	virtual void		train (const Matrix& traindata, int startmonth) {MUST_OVERLOAD}

	/** Tests the data and returns the monthly predictions in matrix.
	 *
	 *  Must be implemented by auditing strategies.
	 *
	 *  @param startmonth See train().
	 **/
	virtual Matrix		predict		(const Matrix& testdata, int startmonth) const {MUST_OVERLOAD; return Matrix(0,0);}

	/** Tests the learning method with the given data.
	 *
	 *  @param inputmonths Most prediction methods require at least one
	 *  month of data before the first data point to be
	 *  predicted. Thus, if you want to predict 12 months sequence,
	 *  you have to give for example 13 months of test data. Some
	 *  prefer two months, while some may require three or even more.
	 *
	 *  @param inputmonths Number of months to be used as prediction
	 *  inputs before the first predicted month. The testdata contains
	 *  these patterns in the beginning in the set.
	 *
	 *  @param startmonth See train().
	 **/
	virtual AuditingTestResults*	test		(const Matrix& testdata, int startmonth) const;

	/** Tests the data and plots the result as a curve with GnuPlot.
	 *
	 *  @param filename File where the EPS pictures should be printed.
	 **/
	virtual void		testCurve	(const Matrix& testdata, int startmonth, const String& filename) const;

	/** Returns the name of the prediction method. */
	const String&		name		() const {return mName;}

	virtual int			inputMonths	() const {return mInputMonths;}
	
  protected:
	String	mName;

	/** Number of months to be used as prediction inputs before the
	 *  first predicted month. The testdata contains these patterns in
	 *  the beginning in the set.  */
	int		mInputMonths;

	static Matrix		rowDeltas			(const Matrix& matrix);
};

class AverageDeltaPrediction : public AuditingStrategy {
	decl_dynamic (AverageDeltaPrediction);
  public:
						AverageDeltaPrediction	() : AuditingStrategy ("AverageDelta") {mInputMonths=1;}
						AverageDeltaPrediction	(const StringMap& params);
	virtual void		train					(const Matrix& traindata, int startmonth);
	virtual Matrix		predict					(const Matrix& testdata, int startmonth) const;
	const Matrix&		deltas					() const {return mMonthAvg;}
	virtual int			inputMonths				() const {return 1;}
  protected:
	Matrix			mMonthAvg;
	PackArray<int>	mMonthCnt;
};

class ZeroDeltaPrediction : public AuditingStrategy {
	decl_dynamic (ZeroDeltaPrediction);
  public:
						ZeroDeltaPrediction		() : AuditingStrategy ("ZeroDelta") {mInputMonths=1;}
						ZeroDeltaPrediction		(const StringMap& params);
	virtual void		train					(const Matrix& traindata, int startmonth);
	virtual Matrix		predict					(const Matrix& testdata, int startmonth) const;
	virtual int			inputMonths				() const {return 1;}
  protected:
};

/** Neural network prediction method.
 *
 *  Very well customizable by the undocumented parameters for the *
 *  make() method.
 **/
class AbsoluteNeuralPrediction : public AuditingStrategy {
	decl_dynamic (AbsoluteNeuralPrediction);
  public:
						AbsoluteNeuralPrediction	(const char* name=NULL) : AuditingStrategy (name? name:"AbsoluteNeural")  {mpNetwork=NULL; rpObserver=NULL;}
						~AbsoluteNeuralPrediction	();
	virtual void		make						(const StringMap& params);
	virtual void		train						(const Matrix& traindata, int startmonth);
	virtual Matrix		predict						(const Matrix& testdata, int startmonth) const;

	void				setObserver					(TrainingObserver* observer) {rpObserver=observer;}
	void				setNetwork					(ANNetwork* net);
	const ANNetwork&	getNetwork					() const {return *mpNetwork;}

  protected:
	/** Trained network. */
	ANNetwork*			mpNetwork;

	/** Flag that says whether we train all variables at a time, i.e.,
	 *  have all the variables in the output layer.
	 **/
	bool				mUseAllOutputs;

	/** Should all input variables be used or just one? If false,
	 *  mUseAllOutputs must also be false (we can't predict all
	 *  variables with the information from just one variable).
	 **/
	bool				mUseAllInputs;

	/** Currently handled output variable. This value changes during
	 *  the run.
	 **/
	int					mVariable;

	/** Should global equalization be used? */
	bool				mGlobalEqualization;

	/** Topology description string for the hidden units, for example
	 *  "10-5-5".
	 **/
	String				mHiddenTopology;

	/** Application parameters, stored here for subsequent use. */
	StringMap			mParams;

	/** Training observer that can print out training log during
	 *  network training.
	 **/
	TrainingObserver*	rpObserver;

  protected:
	/** Builds pattern set from given dataset and starting month. */
	PatternSet*			makeSet					(const Matrix& data, int startmonth) const;

  private:
	int					inputVariables			(int datacolumns) const {return 12+mInputMonths*(mUseAllInputs? datacolumns : 1);}
	int					outputVariables			(int datacolumns) const {return mUseAllOutputs? datacolumns : 1;}
};

class SingleNeuralPrediction : public AbsoluteNeuralPrediction {
	decl_dynamic (SingleNeuralPrediction);
  public:
					SingleNeuralPrediction		() : AbsoluteNeuralPrediction ("SingleNeural")  {mpNetwork=NULL;}
	virtual void	make						(const StringMap& params);
	virtual void	train						(const Matrix& traindata, int startmonth);
	virtual Matrix	predict						(const Matrix& testdata, int startmonth) const;

  protected:
	Array<ANNetwork>	mNetworks;
};

/** Not in use currently.
 *
 **/
class DeltaNeuralPrediction : public AuditingStrategy {
	decl_dynamic (DeltaNeuralPrediction);
  public:
					DeltaNeuralPrediction	() : AuditingStrategy ("DeltaNeural") {}
					DeltaNeuralPrediction	(const StringMap& params);
	virtual void	train					(const Matrix& traindata, int startmonth);
	virtual Matrix	predict					(const Matrix& testdata, int startmonth) const;
  protected:
};

class PreviousYear : public AuditingStrategy {
	decl_dynamic (PreviousYear);
  public:
						PreviousYear			() : AuditingStrategy ("PreviousYear") {mInputMonths=0;}
						PreviousYear			(const StringMap& params);
	virtual void		train					(const Matrix& traindata, int startmonth);
	virtual Matrix		predict					(const Matrix& testdata, int startmonth) const;
	virtual int			inputMonths				() const {return 0;}
  protected:
	Matrix			mData;
};

class PreviousYearsAvg : public AuditingStrategy {
	decl_dynamic (PreviousYearsAvg);
  public:
						PreviousYearsAvg		() : AuditingStrategy ("PreviousYearsAvg") {mInputMonths=0;}
						PreviousYearsAvg		(const StringMap& params);
	virtual void		train					(const Matrix& traindata, int startmonth);
	virtual Matrix		predict					(const Matrix& testdata, int startmonth) const;
	virtual int			inputMonths				() const {return 0;}
  protected:
	Matrix			mMonthlyAvg;
};


///////////////////////////////////////////////////////////////////////////////////////////////////
//   _             | o     o              ----                                     |     o       //
//  / \            |    |      _         (      |       ___   |   ___              |       |     //
// /   \ |   |  ---| | -+- | |/ \   ___   ---  -+- |/\  ___| -+- /   )  ___  \   | |     | |---  //
// |---| |   | (   | |  |  | |   | (   \     )  |  |   (   |  |  |---  (   \  \  | |     | |   ) //
// |   |  \__!  ---| |   \ | |   |  ---/ ___/    \ |    \__|   \  \__   ---/   \_/ |____ | |__/  //
//                                  __/                                 __/   \_/                //
///////////////////////////////////////////////////////////////////////////////////////////////////

class AuditingStrategyLib {
  public:
									AuditingStrategyLib ();
	AuditingStrategy*				create		(int i) const;
	int								strategies	() const {return mClassNames.size;}
	void							registerCls	(const String& classname) {mClassNames.add(classname);}
	static	AuditingStrategyLib&	instance	() {return sInstance;}
  protected:
	static AuditingStrategyLib		sInstance;
	Array<String>					mClassNames;
};

#endif

