/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#include <magic/applic.h>
#include <inanna/patternset.h>
#include <inanna/trainer.h>
#include "auditing.h"

class ReportObserver : public TrainingObserver {
  public:
	void	cycleTrained	(const Trainer& trainer, int totalCycles) {
		printf ("Cycles trained: %d\n", totalCycles);
	}
};

Main () {
	assertmode = ASSERT_CRASH;
	sout.autoflush ();
	readConfig ("auditing.cfg");
		
	//////////////////////////////////////////////////////////////////////
	// PREPARE STRATEGY

	// Create auditing strategy
	AuditingStrategyLib& audStr = AuditingStrategyLib::instance();
	AuditingStrategy* strategy = audStr.create(int(paramMap()["strategy"]));
	strategy->make (paramMap());
	sout.printf ("Strategy: %s\n", (CONSTR) strategy->name());
	if (AbsoluteNeuralPrediction* anp = dynamic_cast<AbsoluteNeuralPrediction*>(strategy))
		anp->setObserver (new ReportObserver ());
	
	//////////////////////////////////////////////////////////////////////
	// PREPARE DATA
	
	PatternSet loaded ("data/inputs.raw", 10, 0);
	fprintf (stderr, "Loaded %d samples of length %d\n", loaded.patterns, loaded.inputs+loaded.outputs);

	// Put data into matrix
	Matrix alldata = loaded.getMatrix ();

	// Split it into training and test set
	Matrix traindata = alldata.sub(1,alldata.rows-13, 1, Matrix::end);
	Matrix testdata = alldata.sub(alldata.rows-12-strategy->inputMonths(), Matrix::end, 1, Matrix::end);
	
	fprintf (stderr, "Train data: %d patterns with %d values\n", traindata.rows, traindata.cols);
	fprintf (stderr, "Test  data: %d patterns with %d values\n", testdata.rows, testdata.cols);

	//////////////////////////////////////////////////////////////////////
	// MAKE PREDICTIONS (train & test)

	// Prepare result holders
	Array<AuditingTestResults> allResults;
	Matrix prediction (12, testdata.cols);
	for (int r=0; r<prediction.rows; r++)
		for (int c=0; c<prediction.cols; c++)
			prediction.get(0,0) = 0.0;

	// Make training-testing runs and average results
	int runs = paramMap()["runs"];
	for (int i=0; i<runs; i++) {
		printf ("Run %d\n", i);
		
		// Train the strategy
		strategy->train (traindata, alldata.get(1,0));
		
		// Test the strategy numerically
		int testmonth = int (alldata.get(alldata.rows-12-strategy->inputMonths(),0));
		AuditingTestResults* results = strategy->test (testdata, testmonth);
		allResults.add (results);
		sout.printf ("Average error: %f\n", results->averageError);
		for (int v=0; v<results->averageErrors.size; v++)
			sout.printf ("Variable %d: %f\n", v, results->averageErrors[v]);
		
		// If ADP, save some minor results
		if (AverageDeltaPrediction* adp = dynamic_cast<AverageDeltaPrediction*>(strategy)) {
			FILE* out = fopen("results/avgDeltas.mat", "w");
			adp->deltas().save (out);
			fclose (out);
		}

		// Make prediction
		prediction += strategy->predict (testdata, testmonth);
	}

	//////////////////////////////////////////////////////////////////////
	// REPORTING
	
	// Calculate averages and stddevs for the results
	Vector averageErrors (allResults.size);								// For all variables
	Array<Vector> averageVarErrors (allResults[0].averageErrors.size);	// For each variable
	for (int i=0; i<averageVarErrors.size; i++)			// Create the vectors
		averageVarErrors[i].make (allResults.size);
	for (int i=0; i<allResults.size; i++) {				// Put the averages into the vectors
		averageErrors[i] = allResults[i].averageError;
		for (int j=0; j<averageVarErrors.size; j++)
			averageVarErrors[j][i] = allResults[i].averageErrors[j];
	}

	// Print statistics
	printf ("Average error for all inputs and runs: %f (stddev %f)\n",
			avg (averageErrors), stddev(averageErrors));
	printf ("Average error for each input for all runs:\n");
	for (int j=0; j<averageVarErrors.size; j++)
		printf ("Variable %d: %f (stddev %f)\n", j, avg (averageVarErrors[j]), stddev (averageVarErrors[j]));

	// Write prediction to a file
	prediction /= runs;
	printf ("Prediction matrix, %dx%d\n", prediction.rows, prediction.cols);
	FILE* out = fopen(format("results/%s.res", (CONSTR)strategy->name()), "w");
	for (int r=0; r<prediction.rows; r++) {
		fprintf (out, "%d ", int (alldata.get(alldata.rows-12+r,0)));
		for (int c=0; c<prediction.cols; c++)
			fprintf (out, "%f ", prediction.get(r,c));
		fprintf (out, "\n");
	}
	fclose (out);
	
	// Finished
	delete strategy;
}


// Barbro: 040-5891428
// http://www.neci.nj.nec.com/homepages/lawrence/papers/finance-tr96/latex.html
// http://www.nodes.de/PredTimeSeries.htm
//
// Jorma Hjort 0400-474747
