#ifndef __ANNETWORK_H__
#define __ANNETWORK_H__

#include <magic/object.h>
#include <magic/Math.h>
#include <magic/cstring.h>
#include <magic/pararr.h>
//#include "trainset.h"

// External
class PatternSet;
class ANNetwork;
class Terminator;

/** Struct for returning classification results from @ref
 *  AnyNetwork::testClassify.
 **/
struct ClassifResults {

	/** Mean squared error of the test. */
	double			mse;

	/** Total number of classification failures. */
	int				failures;

	/** Number of failures for each class. */
	PackArray<int>	classcnts;

	/** Number of instances of each class in the test set. */
	PackArray<int>	classSizes;
};



/////////////////////////////////////////////////////////////////////////////// 
//                                                                           //
//                _   |   | |   |                           |                //
//               / \  |\  | |\  |  ___   |                  |                //
//              /   \ | \ | | \ | /   ) -+- \    /  __  |/\ | /              //
//              |---| |  \| |  \| |---   |   \\//  /  \ |   |/               //
//              |   | |   | |   |  \__    \   VV   \__/ |   | \              //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#define LearningMapping AnyNetwork

/** Abstract interface for any ANN implementations.
 *
 *  This abstraction is intended be generic enough to accomodate every
 *  kind of artificial neural networks. But, since there are currently
 *  just a very few such implementations, this class may not be very
 *  mature.
 *
 *  Slightly bloated.
 **/
class LearningMapping : public Object {
  public:

							LearningMapping		();
	virtual					~LearningMapping		() {;}
	
	/** Creates a network according to the given structure
	 *  description.  The exact interpretation of the description are
	 *  dependent on the implementor.
	 **/
	virtual void			make			(const char* description) = 0;

	/** Initializes the network with the given weight range (use
	 *  default if negative).
	 **/
	virtual void			init			(double r=-1) = 0;

	/** Trains the network. MUST BE OVERLOADED if the ANN system uses
	 *  any training (not necessary if it doesn't).
	 *
	 *  @return Final MSE (mean squared error) at the end of the training.
	 *
	 *  @param set Training set.
	 *
	 *  @param cycles Maximum number of training cycles. The training
	 *  may be terminated before this by the @ref Terminator method
	 *  used.
	 *
	 *  @param cycint Validation interval for @ref Terminator (early
	 *  stopping) method. Default value -1 disables validation and
	 *  thus early stopping, and training is carried out for the
	 *  maximum number of training cycles.
	 *
	 *  @param vsize OPTIONAL Number of patterns in the 'set' used as
	 *  a validation set.
	 **/
	virtual double			train			(const PatternSet& set, int cycles,
											 int cycint=-1, int vsize=-1) {MUST_OVERLOAD; return 0.0;}

	/** Trains the network. Coating for the other train method above;
	 *  accepts validation set as a parameter.
	 *
	 *  @param trainset Training set.
	 *  @param validationSet Validation set.
	 *  @param cycles Maximum number of training cycles. The training
	 *  may be terminated before this by the @ref Terminator method
	 *  used.
	 *
	 *  @param cycint Validation interval for @ref Terminator (early
	 *  stopping) method. Default value -1 disables validation and
	 *  thus early stopping, and training is carried out for the
	 *  maximum number of training cycles.
	 **/
	double					train			(const PatternSet& trainSet,
											 const PatternSet& validationSet,
											 int cycles, int cycint=-1);

	/** Trains the network for one cycle. MUST BE OVERLOADED if
	 *  training is desired.
	 **/
	virtual double			trainOnce		(const PatternSet& trainset) {MUST_OVERLOAD return 0.0;}
	
	/** Tests a specific pattern from a set. Returns a double vector
	 *  containing the output values.
	 *
	 *  @return Vector containing the values of output units.
	 *
	 *  @param set The @ref PatternSet where the pattern is stored.
	 *  @param pattern Index number of the pattern in the @ref PatternSet.
	 **/
	virtual Vector			testPattern		(const PatternSet& set, int pattern) const {MUST_OVERLOAD; return 0.0;}

	// These should not be overridden usually
	
	/** Tests an entire pattern set.
	 *
	 *  @return Returns MSE (mean squared error).
	 **/
	virtual double			test			(const PatternSet& set) const;

	/** Makes a classification test with the given test set. If there
	 *  is just one output unit, it is interpreted to indicate two
	 *  class memberships. In any other case, each output unit
	 *  designates different class.
	 *
	 *  @return Returns the classification results for each class in a
	 *  @ref ClassifResults struct.
	 **/
	virtual ClassifResults*	testClassify	(const PatternSet& set) const;

	// Common conversions

	/** Copy or conversion from @ref ANNetwork. Implementing this is
	 *  essential for the use of @ref Terminator early stopping,
	 *  because transient network states are stored in ANNetwork
	 *  objects during training.
	 *
	 *  @param onlyWeights If this flag is 'true', both networks are
	 *  assumed to have the same topology. Therefore, only weights and
	 *  activations have to be copied.
	 **/
	virtual void			copyFreeNet		(const ANNetwork& fnet,
											 bool onlyWeights=false);

	/** Copy or conversion from any other neural network.
	 *
	 *  @param fnet LearningMapping to be copied. It is probable that the
	 *  implementor wants to use dynamic_cast to find out the true
	 *  class of the source class.
	 *
	 *  @param onlyWeights If this flag is 'true', both networks are
	 *  assumed to have the same topology. Therefore, only weights and
	 *  activations have to be copied.
	 **/
	virtual void			copy			(const LearningMapping& fnet,
											 bool onlyWeights=false);

	/** Conversion to @ref ANNetwork, a highly object-oriented
	 *  network representation.
	 **/
	virtual ANNetwork*	toANNetwork	() const {MUST_OVERLOAD; return NULL;}

};


#endif

