/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __FILEFORMAT_H__
#define __FILEFORMAT_H__

#include "freenet.h"

class ANNFileFormat; // Internal

EXCEPTIONCLASS (invalid_filename);



////////////////////////////////////////////////////////////////////////////////////
//   _   |   | |   | ----- o |       -----                          |     o       //
//  / \  |\  | |\  | |       |  ___  |                     ___   |  |       |     //
// /   \ | \ | | \ | |---  | | /   ) |---   __  |/\ |/|/|  ___| -+- |     | |---  //
// |---| |  \| |  \| |     | | |---  |     /  \ |   | | | (   |  |  |     | |   ) //
// |   | |   | |   | |     | |  \__  |     \__/ |   | | |  \__|   \ |____ | |__/  //
////////////////////////////////////////////////////////////////////////////////////

/** ANNetwork data file format library.
 *
 *  Patterns:
 *   - Singleton, Abstract Factory.
 **/
class ANNFileFormatLib {
  public:

	/** Creates a network from the given file.
	 *
	 *  @throws file_not_found, invalid_format, assertion_failed
	 **/
	static ANNetwork*	load	(const String& filename) throw (file_not_found, invalid_format, assertion_failed);
	
	/** Tries to load the given file to the given network.
	 *
	 *  @throws file_not_found, invalid_format, assertion_failed
	 **/
	static void		load	(const String& filename, ANNetwork& net) throw (file_not_found, invalid_format, assertion_failed);

	/** Loads network from the given stream.
	 *
	 *  @param in Input stream
	 *
	 *  @param filetype File name extension; ".raw",".pat" (SNNS), or ".dt" (Proben1)
	 *
	 *  @throws file_not_found, invalid_format, assertion_failed
	 **/
	static void		load	(istream& in, ANNetwork& net, const String& filetype) throw (invalid_format, assertion_failed);
	
	/** Writes the given network to file with the given name. File
	 *  format can also be specified.
	 *
	 *  @throws file_not_found, invalid_format, assertion_failed
	 **/
	static void		save	(const String& filename, const ANNetwork& set, const char* fileformat=NULL) throw (invalid_filename, invalid_format, assertion_failed);

  protected:
	/** Factory creates a data format handler according to
	 *  filename. To make the factory more extensible, there would
	 *  need to be some sort of dynamic registry of Factory Methods.
	 **/
	static ANNFileFormat*	create	(const String& fileformat) throw (invalid_format);
};



///////////////////////////////////////////////////////////////////////
//   _   |   | |   | ----- o |       -----                           //
//  / \  |\  | |\  | |       |  ___  |                     ___   |   //
// /   \ | \ | | \ | |---  | | /   ) |---   __  |/\ |/|/|  ___| -+-  //
// |---| |  \| |  \| |     | | |---  |     /  \ |   | | | (   |  |   //
// |   | |   | |   | |     | |  \__  |     \__/ |   | | |  \__|   \  //
///////////////////////////////////////////////////////////////////////

/** Baseclass for @ref ANNetwork file formats.
 *
 *  Patterns:
 *   - Strategy or
 *   - Template Method (subclasses implement stream I/O),
 **/
class ANNFileFormat {
  public:

	/** Tries to load the given file to the given pattern set.
	 *
	 *  @throws invalid_format, assertion_failed
	 **/
	virtual void	load	(istream& in, ANNetwork& net) const=0;
	virtual void	save	(ostream& out, const ANNetwork& net) const {MUST_OVERLOAD}

  protected:
};

class SNNS_ANNFormat : public ANNFileFormat {
  public:
	virtual void	load	(istream& in, ANNetwork& set) const;
	virtual void	save	(ostream& out, const ANNetwork& set) const;
};


#endif
