/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#include <magic/object.h>
#include <magic/list.h>
#include <magic/map.h>
#include <magic/RegExp.h>

#include "dataformat.h"
#include "dataformats.h"


//////////////////////////////////////////////////////////////////////////////////
//  ---- |   | |   |  ---- ___                   -----                          //
// (     |\  | |\  | (     |  \   ___   |   ___  |                     ___   |  //
//  ---  | \ | | \ |  ---  |   |  ___| -+-  ___| |---   __  |/\ |/|/|  ___| -+- //
//     ) |  \| |  \|     ) |   | (   |  |  (   | |     /  \ |   | | | (   |  |  //
// ___/  |   | |   | ___/  |__/   \__|   \  \__| |     \__/ |   | | |  \__|   \ //
//////////////////////////////////////////////////////////////////////////////////

struct PatternRow {
	Vector	values;
	String	comment;
	PatternRow (int n) {values.make(n);}
};

void SNNSDataFormat::load (istream& in, PatternSet& set) const {

	// Some nice regular expressions to extract data from SNNS info lines
	RegExp no_patts	("patterns : (.+)");
	RegExp no_ins	("input units : (.+)");
	RegExp no_outs	("output units : (.+)");
	RegExp inpatt	("Input pattern (.+):");
	RegExp outpatt	("Output pattern (.+):");
	
	// State variables
	enum states {ST_NONE=0,
				 ST_INS=1,
				 ST_OUTS=2,
				 ST_NONE_INS=3,
				 ST_NONE_OUTS=4
				};
	int				state	= ST_NONE;	// Nice two bit flip-flop state
	int				readcnt	=0;			// Amount of values in readarr
	String			buff;				// Input buffer
	Array<String>	linesubs;			// Subexpressions
	int				pattern	= 0;		// Pattern counter: current pattern
	int				patts	= 0;
	int				ins		= 0;
	int				outs	= 0;
	
	while (in >> buff) {

		if (isdigit (buff[0])) {
			if (!set.inputs || !set.outputs || !set.patterns)
				throw invalid_format ("Pattern file dimensions not given anywhere");

			// An interestingly strange state machine. I wonder how it
			// works. This is propably a good example of my
			// "evolutionary programming style", which means trying
			// until it works and not thinking. Or, maybe, this time I
			// superceded my thinking ability; my subconscious
			// intelligence must be enormous.
			switch (state) { // PLEASE NOTE THE CUTE FALLTHROUGHS! (They are really significant)
			  case ST_NONE_OUTS:
				  if (pattern>0 && readcnt!=set.outputs)
					  throw invalid_format (format ("Too short output vector #d in SNNS "
													"pattern set", pattern));
				  pattern++;
			  case ST_NONE:
				  state = ST_INS;
				  readcnt=0;
			  case ST_INS:
				  if (readcnt<set.inputs)
					  set.set_input (pattern, readcnt++, double(String (buff)));
				  break;
			  case ST_NONE_INS:
				  if (pattern>0 && readcnt!=set.inputs)
					  throw invalid_format (format ("Too short input vector #d in SNNS "
													"pattern set", pattern));
				  state = ST_OUTS;
				  readcnt = 0;
			  case ST_OUTS:
				  if (readcnt<set.outputs)
					  set.set_output (pattern, readcnt++, double (String (buff)));
			}
			
		} else {
			// The line seems to be text, read the rest of the line
			getS (in, buff);

			// Extract number of patterns
			if (no_patts.match (buff, linesubs)) {
				patts = linesubs[1];
				if (set.patterns && patts!=set.patterns)
					throw invalid_format ("Pattern file has wrong dimensions");
			}

			// Extract number of inputs
			if (no_ins.match (buff, linesubs)) {
				ins = linesubs[1];
				if (set.inputs && ins!=set.inputs)
					throw invalid_format ("Pattern file has wrong dimensions");
			}

			// Extract number of outputs
			if (no_outs.match (buff, linesubs)) {
				outs = linesubs[1];
				if (set.outputs && outs!=set.outputs)
					throw invalid_format ("Pattern file has wrong dimensions");
			}

			// If we have read the set dimensions from the file, but
			// the set has not yet dimensions, create it.
			if ((!set.inputs || !set.outputs) && ins && outs && set.patterns)
				set.make (patts, ins, outs);

			// Jump to the expect states
			if (state==ST_INS)
				state = ST_NONE_INS;
			if (state==ST_OUTS)
				state = ST_NONE_OUTS;
		}
	}

	if (pattern!=set.patterns-1)
		throw invalid_format (format ("Wrong number of patterns in SNNS pattern file "
									  "(found %d of %d", pattern, set.patterns));
}

void SNNSDataFormat::save (FILE* out, const PatternSet& set) const {
	fprintf (out, "SNNS pattern definition file V3.2\n"
			 "generated at Xxxxx time\n\n\n"
			 "No. of patterns : %d\n"
			 "No. of input units : %d\n"
			 "No. of output units : %d\n\n",
			 set.patterns, set.inputs, set.outputs);

	for (int p=0; p<set.patterns; p++) {
		fprintf (out, "# input %d\n", p);
		for (int i=0; i<set.inputs; i++) {
			if (i>0)
				fprintf (out, " ");
			if (is_undef (set.input (p,i)))
				fprintf (out, "0");
			else 
				fprintf (out, "%g", set.input (p,i));
		}
		fprintf (out, "\n# target %d\n", p);
		for (int i=0; i<set.outputs; i++) {
			fprintf (out, " ");
			if (is_undef (set.output (p,i)))
				fprintf (out, "0");
			else 
				fprintf (out, "%g", set.output (p,i));
		}
		fprintf (out, "\n");
	}
}


///////////////////////////////////////////////////////////////////////////////////////////
// ----                             ___                   -----                          //
// |   )          |      ___    _   |  \   ___   |   ___  |                     ___   |  //
// |---  |/\  __  |---  /   ) |/ \  |   |  ___| -+-  ___| |---   __  |/\ |/|/|  ___| -+- //
// |     |   /  \ |   ) |---  |   | |   | (   |  |  (   | |     /  \ |   | | | (   |  |  //
// |     |   \__/ |__/   \__  |   | |__/   \__|   \  \__| |     \__/ |   | | |  \__|   \ //
///////////////////////////////////////////////////////////////////////////////////////////

void Proben1DataFormat::load (istream& in, PatternSet& set) const {
	String buff;	// Read buffer
	
	// Read header
	String header;
	header.reserve (256);
	for (int i=0; i<7; i++) {
		getS(in, buff);
		buff.chop();
		ASSERT (buff.len>0);
		ASSERT (buff.find ("=")!=-1);
		header += buff;
		header += '&';
	}

	// Split the header into a map
	StringMap map;
	splitpairs (map, header, '=', '&');

	// Check that all required fields exist
	bool kludge=false;
	if (isnull(map["bool_in"])) kludge=true;
	if (isnull(map["real_in"])) kludge=true;
	if (isnull(map["bool_out"])) kludge=true;
	if (isnull(map["real_out"])) kludge=true;
	if (isnull(map["training_examples"])) kludge=true;
	if (isnull(map["validation_examples"])) kludge=true;
	if (isnull(map["test_examples"])) kludge=true;
	if (kludge)
		throw generic_exception ("Some required line(s) were missing from a "
								 "Proben1 data file header");

	// Extract the number of patterns, inputs and outputs
	int ins  = int(map["bool_in"]) + int(map["real_in"]);
	int outs = int(map["bool_out"]) + int(map["real_out"]);
	int pats = int(map["training_examples"]) + int(map["validation_examples"])
		+ int(map["test_examples"]);

	// Make my day. Oops. ^H^H^H^H^H^H^H^H^Hmyself
	set.make (pats, ins, outs);

	// Read the patterns
	for (int p=0; p<pats; p++) {
		// Read one line
		getS (in, buff);
		buff.chop ();

		// Split it to fields (MUST be delimited with one space)
		Array<String> items;
		buff.split (items, ' ');

		// Count the number of items
		int pos=0;
		Array<String> fields (items.size);
		for (int i=0; i<items.size; i++)
			if (items[i].len>0)
				fields[pos++] = items[i];
		fields.resize (pos);
		
		ASSERTWITH (fields.size==set.inputs+set.outputs,
					format ("Wrong number of fields (%d) in a Proben1 file "
							"(%d was expected).\n"
							"Propably due to bad formatting in line:\n"
							"%s\n"
							"Rules: number of fields must be equal to "
							"the sum of inputs and outputs and they \n"
							"must be separated by spaces (ASC32). No "
							"other characters are allowed\n",
							fields.size, set.inputs+set.outputs, (CONSTR) buff));

		// Extract input values
		for (int i=0; i<ins; i++)
			set.set_input (p,i, double(fields[i]));

		// Extract output values
		for (int i=0; i<outs; i++)
			set.set_output (p,i, double(fields[i+ins]));
	}

	// Store sizes to dynamic attributes
	set.setAttribute ("training_examples", new Int (int(map["training_examples"])));
	set.setAttribute ("validation_examples", new Int (int(map["validation_examples"])));
	set.setAttribute ("test_examples", new Int (int(map["test_examples"])));
}



///////////////////////////////////////////////////////////////////////////////
//  ----               ___                   -----                           //
//  |   )  ___         |  \   ___   |   ___  |                     ___   |   //
//  |---   ___| \    / |   |  ___| -+-  ___| |---   __  |/\ |/|/|  ___| -+-  //
//  | \   (   |  \\//  |   | (   |  |  (   | |     /  \ |   | | | (   |  |   //
//  |  \   \__|   VV   |__/   \__|   \  \__| |     \__/ |   | | |  \__|   \  //
///////////////////////////////////////////////////////////////////////////////

void RawDataFormat::load (istream& in, PatternSet& set) const {
	List<PatternRow*> patlist;	// Will contain the rows in the input file
	int patNum=0;				// Number of patterns in the input file
	
	// Read the datafile
	String buffer;
	while (getS (in, buffer, '\n')) { // Read one line until EOF
		PatternRow* prow = new PatternRow (set.inputs? set.inputs+set.outputs : 1);
		String item;		// One numeric or text item
		int itemIndex=0;	// Index of the above item within the row

		// Parse the row
		for (int i=0; i<buffer.len; i++) {
			// If we have a field break, or End-Of-Line
			if (isspace (buffer[i])) {
				// If we don't know the number of columns, we must
				// grow it by one
				if (set.inputs==0)
					prow->values.resize (itemIndex+1);

				// Copy parsed value to row vector
				if (set.inputs==0 || itemIndex<set.inputs+set.outputs)
					prow->values[itemIndex] = (item=="x")? UNDEFINED_FLOAT : double(item);
				item="";

				// Check if we have read enough of numeric items
				itemIndex++;
				if (set.inputs>0 && itemIndex >= set.inputs+set.outputs) {
					prow->comment = buffer.substr (i);
					break;
				}
			} else
				item += buffer[i];
		}

		// If we didn't know the number of inputs before, we know it
		// after reading the first row.
		if (set.inputs==0 && itemIndex>0)
			set.inputs = itemIndex;

		// Add the row only if it had enough fields
		if (itemIndex==set.inputs)
			patlist.add (prow);

		patNum++;
	}

	// Put the pattern list into the set

	// Adjust our size
	set.make (patNum, set.inputs, set.outputs);

	Array<String>* comments = new Array<String> (patNum);

	int p=0;
	for (ListIter<PatternRow*> l (patlist); !l.exhausted(); l.next(), p++) {
		for (int i=0; i<set.inputs; i++)
			set.set_input (p, i, l.get()->values[i]);
		for (int i=0; i<set.outputs; i++)
			set.set_output (p, i, l.get()->values[i+set.inputs]);
		(*comments)[p] = l.get()->comment;
	}

	set.setAttribute ("comments", comments);
}
 
void RawDataFormat::save (FILE* out, const PatternSet& set) const {
	for (int p=0; p<set.patterns; p++) {
		for (int i=0; i<set.inputs; i++) {
			if (i>0)
				fprintf (out, " ");
			if (is_undef (set.input (p,i)))
				fprintf (out, "x");
			else 
				fprintf (out, "%g", set.input (p,i));
		}
		for (int i=0; i<set.outputs; i++) {
			fprintf (out, " ");
			if (is_undef (set.output (p,i)))
				fprintf (out, "x");
			else 
				fprintf (out, "%g", set.output (p,i));
		}
		if (!isnull(set.getAttribute("comments")))
			fprintf (out, " %s", (CONSTR) dynamic_cast<const Array<String>&> (set.getAttribute("comments"))[p]);
		fprintf (out, "\n");
	}
}

