/***************************************************************************
 *   copyright            : (C) 2000 by Marko Grnroos
 *   email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#ifndef __EQUALIZER_H__
#define __EQUALIZER_H__

#include <magic/object.h>
#include <magic/Math.h>
#include <magic/Matrix.h>
#include <magic/pararr.h>

// XML format ios flag
extern int xmlflag;



//////////////////////////////////////////////////////////////////////////////
//                 -----                  | o                               //
//                 |                 ___  |   ___  ___                      //
//                 |---   __  |   |  ___| | |   / /   ) |/\                 //
//                 |     /  \ |   | (   | | |  /  |---  |                   //
//                 |____ \__|  \__!  \__| | | /__  \__  |                   //
//                          |                                               //
//////////////////////////////////////////////////////////////////////////////

/** Abstract interface for vector equalization methods. We do not
 *  actually equalize pattern sets, but more generally single vectors
 *  which each are one column (a component or variable plane) of a
 *  pattern set matrix.
 *
 *  You can equalize entire pattern sets by using the MatrixEqualizer.
 **/
class Equalizer : public Object {
	decl_dynamic (Equalizer);
  public:
						Equalizer		() {mHandleMissing=false;}
						Equalizer		(const Equalizer& o) {mHandleMissing=o.mHandleMissing;}
	
	/** Analyzes the vector and calculates equalization parameters for
	 *  it. Equalization is applied by the equalize() method below.
	 *
	 * @param vec Vector to be analyzed for equalization.
	 *
	 * @param additive Should the analysis be additive? If true, all
	 * data is analyzed in accumulating manner. Additive analysis is
	 * identical to concatenating multiple vectors and then calling
	 * analyze(vec,false).
	 **/
	virtual void		analyze			(const Vector& vec, bool additive=false) {MUST_OVERLOAD}

	/** Applies equalization to the given vector.
	 *
	 * Note for implementors: handle missing parameters, if enabled by
	 * mHandleMissing setting.
	 **/
	virtual void		equalize		(Vector& vec) const  {MUST_OVERLOAD}

	/** Applies equalization reversely to the given vector.
	 **/
	virtual void		unequalize		(Vector& vec) const {MUST_OVERLOAD}

	/** Standard clone operator, which must be implemented by
	 *  inheritors. It is required by the MatrixEqualizer to create
	 *  copies of equalizer objects.
	 **/
	virtual Equalizer*	clone			() const  {MUST_OVERLOAD; return NULL;}

	/** Implementation of serialization. */
	virtual	ostream&	operator>>		(ostream&) const {MUST_OVERLOAD; return cout;}

	/** Sets whether or not the equalization should handle missing
	 *  values in a standard manner.
	 **/
	virtual void		handleMissing	(bool enable=true) {mHandleMissing=enable;}

  protected:
	bool	mHandleMissing;
};



///////////////////////////////////////////////////////////////////////////////
//          |   | o                                      -----               //
//          |   |    ____  |                  ___        |                   //
//          |---| | (     -+-  __   ___  |/\  ___| |/|/| |---   __           //
//          |   | |  \__   |  /  \ (   \ |   (   | | | | |     /  \          //
//          |   | | ____)   \ \__/  ---/ |    \__| | | | |____ \__|          //
//                                  __/                           |          //
///////////////////////////////////////////////////////////////////////////////

/** Histogram equalization method (strategy).
 **/
class HistogramEq : public Equalizer {
	decl_dynamic (HistogramEq);
  public:
							HistogramEq	(int precision=100000, double floor=0.0, double ceiling=1.0);
							HistogramEq (const HistogramEq& orig);
	virtual HistogramEq*	clone		() const {return new HistogramEq (*this);}
	
	virtual void			analyze		(const Vector& vec, bool additive=false);
	virtual void			equalize	(Vector& vec) const;
	virtual void			unequalize	(Vector& vec) const;
	
	/** Implementation of serialization. */
	virtual	ostream&	operator>>	(ostream&) const;

  private:
	static int		determineDomain (float x, int precision, float lowBound, float upBound);
	void			finalize		();
	void			analyze2		(const Vector& vec);

	double			mLowBound;	// Bound in the unequalized data
	double			mUpBound;	// Bound in the unequalized data
	double			mMin;		// Desired lower bound in the equalized data
	double			mMax;		// Desired upper bound in the equalized data
	Array<Vector>	mData;		// Accumulated additive analysis data
	Vector			mHistogram;	// Histogram formed by analysis
};



//////////////////////////////////////////////////////////////////////////////
//           ----                         o             -----               //
//          |      ___         ____  ____    ___    _   |                   //
//          | ---  ___| |   | (     (     |  ___| |/ \  |---   __           //
//          |   \ (   | |   |  \__   \__  | (   | |   | |     /  \          //
//          |___/  \__|  \__! ____) ____) |  \__| |   | |____ \__|          //
//                                                               |          //
//////////////////////////////////////////////////////////////////////////////

/** Normal distribution equalization method (strategy).
 **/
class GaussianEq : public Equalizer {
	decl_dynamic (GaussianEq);
  public:
						GaussianEq	(double avg=0.0, double stddev=0.0) {mAverage=0; mStdDev=0;}
						GaussianEq	(const GaussianEq& orig);
	virtual GaussianEq*	clone		() const {return new GaussianEq (*this);}
	
	virtual void		analyze		(const Vector& vec, bool additive=false);
	virtual void		equalize	(Vector& vec) const;
	virtual void		unequalize	(Vector& vec) const;

  private:
	double	mAverage;
	double	mStdDev;
};



//////////////////////////////////////////////////////////////////////////////
//                 |   | o                       -----                      //
//                 |\ /|     _          ___      |                          //
//                 | V | | |/ \  |/|/|  ___| \ / |---   __                  //
//                 | | | | |   | | | | (   |  X  |     /  \                 //
//                 |   | | |   | | | |  \__| / \ |____ \__|                 //
//                                                        |                 //
//////////////////////////////////////////////////////////////////////////////

/** Linear range equalization method (strategy).
 **/
class MinmaxEq : public Equalizer {
	decl_dynamic (MinmaxEq);
  public:
						MinmaxEq	(double trgMin=0.0, double trgMax=1.0);
						MinmaxEq	(const MinmaxEq& orig);
	virtual MinmaxEq*	clone		() const {return new MinmaxEq (*this);}
	
	virtual void		analyze		(const Vector& vec, bool additive=false);
	virtual void		equalize	(Vector& vec) const;
	virtual void		unequalize	(Vector& vec) const;
	
	/** Implementation of serialization. */
	virtual	ostream&	operator>>	(ostream& out) const;

	/** Implementation of serialization. */
	virtual	istream&	operator<<	(istream& in);
	
  private:
	double	mDataMin;
	double	mDataMax;
	double	mTrgMin;
	double	mTrgMax;
};



//////////////////////////////////////////////////////////////////////////////
//    |   |               o     -----                  | o                  //
//    |\ /|  ___   |            |                 ___  |   ___  ___         //
//    | V |  ___| -+- |/\ | \ / |---   __  |   |  ___| | |   / /   ) |/\    //
//    | | | (   |  |  |   |  X  |     /  \ |   | (   | | |  /  |---  |      //
//    |   |  \__|   \ |   | / \ |____ \__|  \__!  \__| | | /__  \__  |      //
//                                       |                                  //
//////////////////////////////////////////////////////////////////////////////

/** Equalizes matrices. Stores equalization information. */
class MatrixEqualizer : public Equalizer {
	decl_dynamic (MatrixEqualizer);
  public:

	/** Constructor. The prototype is an Equalizer instance which has
	 *  been created with appropriate parameters for the specific
	 *  equalization method.
	 *
	 *  The prototype is applied for building equalizers for each
	 *  column plane.
	 *
	 *  The MatrixEqualizer takes the ownership of the prototype and
	 *  destroys it accordingly.
	 **/
						MatrixEqualizer		(Equalizer* prototype=NULL);

	/** Analyzes a matrix for equalization.
	 *
	 *  If the optional additive-parameter is true, the analysis is
	 *  made for all columns at the same time, i.e., they share their
	 *  value space. This may be useful if the relative values of
	 *  different columns are significant.
	 **/
	void				analyze				(const Matrix& mat, bool additive=false);

	/** Equalizes a matrix by applying equalization data acquired from
	 *  a previous analyze().
	 **/
	void				equalize			(Matrix& mat) const;

	/** Reverses the equalization of a matrix by applying equalization
	 *  data acquired from a previous analyze().
	 **/
	void				unequalize			(Matrix& mat) const;

	/** Returns equalizer for given matrix column (component
     *  plane). The equalizer may have analyzed data or it may have
     *  not, depending on its history.
	 **/
	const Equalizer&	getPlane			(int plane) const {return *mPlaneEqualizers.getp(plane);}

	/** Returns number of columns (equalization planes). */
	int					planes				() const {return mPlaneEqualizers.size;}
	
	/** Implementation of serialization. */
	virtual	ostream&	operator>>	(ostream& out) const;

	/** Implementation of serialization. */
	virtual	istream&	operator<<	(istream& in);

	virtual void		handleMissing	(bool enable=true);
	
  private:
	void				bothEqualize		(Matrix& mat, bool uneq) const;

	/** Equalizers for each column plane for analyzed matrices. */
	Array<Equalizer>	mPlaneEqualizers;
};

////////////////////////////////////////////////////////////////////////////////

/** Dynamically creates and reads an equalizer object from the given
 *  input stream.
 **/
Equalizer* readEqualizer (istream& in);

#endif
