#ifndef __INITIALIZER_H__
#define __INITIALIZER_H__

#include<magic/object.h>
#include"freenet.h"

/** Abstract interface class for @ref Neuron weight initializers.
 **/
class NeuronInitializer : public Object {
  public:
					NeuronInitializer	() {}
					NeuronInitializer	(const NeuronInitializer& orig) {}

	/** The implementor should overload this to initialize the neuron
	 *  and it's weights.
	 **/
	virtual void	initialize			(Neuron& neuron)=0;

	/** Implementor has to overload this if the initializers are to be
	 *  cloned.
	 **/
	virtual NeuronInitializer*	clone	() const {MUST_OVERLOAD}

  private:
	operator= (const NeuronInitializer& orig) {FORBIDDEN}
	decl_dynamic (NeuronInitializer);
};

/** Standard initializer for @ref Neuron weights.
 **/
class GaussianInitializer : public NeuronInitializer {
  public:

	/** Standard constructor.
	 *
	 *  @param r Standard deviation for weight initialization.
	 **/
					GaussianInitializer	(double r=0.5) : mR (r) {}
					GaussianInitializer (const GaussianInitializer& orig) : mR (orig.mR) {}

	/** Initializes the neuron's weights with standard, normally
	 *  distributed initialization function.
	 **/
	void			initialize			(Neuron& neuron) {
		neuron.init (mR);
	}
	
	/** Implementation for @ref Object. */
	virtual GaussianInitializer* clone	() const {new GaussianInitializer (*this);}
	
  private:
	double	mR;

	operator= (const GaussianInitializer& orig) {FORBIDDEN}
	decl_dynamic (GaussianInitializer);
};

/** Dummy @ref NeuronInitializer just leaves the neurons uninitialized.
 **/
class DummyInitializer : public NeuronInitializer {
  public:
					DummyInitializer	() {}
					DummyInitializer	(const DummyInitializer& orig) {}

	/** Dummy. Does not initialize.
	 **/
	void			initialize			(Neuron& neuron) {}

	/** Implementation for @ref Object. */
	virtual DummyInitializer* clone		() const {new DummyInitializer (*this);}
  private:

	operator= (const DummyInitializer& orig) {FORBIDDEN}
	decl_dynamic (DummyInitializer);
};

#endif
