/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#include <magic/object.h>
#include <magic/Math.h>
#include <magic/map.h>
#include <magic/Class.h>
#include <stdio.h>
#include <ctype.h>
#include "patternset.h"
#include "dataformat.h"


// #define rnd(range) (rand()%range)

impl_abstract (PatternSource, {CObject});

PatternSource::PatternSource () {
	patterns	= 0;
	inputs		= 0;
	outputs		= 0;
}

PatternSource::PatternSource (const PatternSource& orig) {
	patterns	= orig.patterns;
	inputs		= orig.inputs;
	outputs		= orig.outputs;
	mName		= orig.mName;
}

void PatternSource::make1 (int patts, int ins, int outs) {
	patterns	= patts;
	inputs		= ins;
	outputs		= outs;
}

int PatternSource::getClass (int p) const {
	if (outputs>1) {
		// Multiple class indicators -> scan trough them to find the highest
		double max=-666;
		int maxI=0;
		for (int i=0; i<outputs; i++)
			if (output(p,i) > max) {
				max = output(p,i);
				maxI = i;
			}
		return maxI;
	} else {
		// Single, binary class indicator
		return (output(p,0)>0.5)? 1:0;
	}
}

void PatternSource::copy (const PatternSource& other, int startp, int endp) {
	ASSERTWITH (startp<other.patterns && endp<other.patterns || endp==startp-1,
				format ("Copy range (%d-%d) out of source set limits (%d patterns)",
						startp, endp, other.patterns));
	ASSERTWITH (startp>=-1, format ("Invalid copy range %d", startp));

	// If the range is empty, do nothing
	if (endp==startp-1)
		return;
	
	ASSERTWITH (startp<=endp, "Start pattern must be greater than end pattern. "
				"Note: end point can be start-1, in which case no copying is done.");

	// Set parameters to default of not given
	if (startp==-1)
		startp = 0;
	if (endp==-1)
		endp = other.patterns-1;

	// Resize self
	make (endp-startp+1, other.inputs, other.outputs);

	// Copy
	for (int p=0; p<=endp-startp; p++) {
		for (int i=0; i<inputs; i++)
			set_input (p, i, other.input (p+startp, i));
		for (int i=0; i<outputs; i++)
			set_output (p, i, other.output (p+startp, i));
	}
}

void PatternSource::join (const PatternSource& a, const PatternSource& b) {
	// First check the possibility that either or both of the sets might be empty
	if (a.patterns==0 || b.patterns==0) {
		// These cases we handle just by copying. For some reason.
		make (0,0,0);
		if (a.patterns>0)
			copy (a);
		if (b.patterns>0)
			copy (b);
		return;
	}

	// Not empty
	
	ASSERTWITH (a.inputs==b.inputs && a.outputs==b.outputs,
				format ("Tsets to be joined must be of equal dimensions (was: %d->%d . %d->%d)",
						a.inputs, a.outputs, b.inputs, b.outputs));
	ASSERTWITH (a.patterns==0 || b.patterns==0 ||
				a.inputs>0 && a.outputs>0 && b.inputs>0 && b.outputs>0,
				"Tsets to be joined may not have null dimension");
	make (a.patterns+b.patterns, a.inputs, a.outputs);

	// Copy a
	for (int p=0; p<a.patterns; p++) {
		for (int i=0; i<inputs; i++)
			set_input (p, i, a.input (p, i));
		for (int i=0; i<outputs; i++)
			set_output (p, i, a.output (p, i));
	}

	// Copy b
	for (int p=0; p<b.patterns; p++) {
		for (int i=0; i<inputs; i++)
			set_input (p+a.patterns, i, b.input (p, i));
		for (int i=0; i<outputs; i++)
			set_output (p+a.patterns, i, b.output (p, i));
	}
}

void PatternSource::split (PatternSource& a, PatternSource& b, double ratio) const {
	ASSERT (ratio>=0 && ratio<=1);
	ASSERT (patterns>0);
	ASSERT (inputs>0);

	a.make (int(patterns*ratio), inputs, outputs);
	b.make (patterns-int(patterns*ratio), inputs, outputs);

	a.copy (*this, 0, int(patterns*ratio)-1);
	b.copy (*this, int(patterns*ratio), patterns-1);

	/*
	ASSERTWITH (a.patterns>0 && b.patterns>0,
				"Resulting pattern set was empty (ratio was maybe too near 0.0 or 1.0,\n"
				"    or the source set was too small?)");
				*/
}

void PatternSource::filter (const PatternSource& src, const String& bits) {
	ASSERTWITH (src.patterns>0 && src.inputs+src.outputs>0,
				"Filter source TSet must have dimensions");
	ASSERTWITH (src.inputs == bits.len || bits.len==0,
				format ("Filter mask must have length (was %d) "
						"equal to source TSet input vector dimension (was %d)",
						bits.len, src.inputs));
	
	// Count features
	int features=0;
	if (bits.len==0)
		features = src.inputs;
	else
		for (int i=0; i<bits.len; i++)
			if (bits[i] == '1')
				features++;
	
	ASSERTWITH (features>0, "Training set filter may not be empty");

	// Re-make self to this size
	make (src.patterns, features, src.outputs);

	// For each pattern
	for (int p=0; p<patterns; p++) {
		
		// Copy the input features by filtering
		int feature=0;
		for (int i=0; i<src.inputs; i++)
			if (bits.len==0 || bits[i]=='1')
				set_input (p, feature++, src.input (p, i));
		
		// Copy the outputs
		for (int j=0; j<src.outputs; j++)
			set_output (p, j, src.output (p, j));
	}
}


Array<int> PatternSource::countClasses () const return result {
	result.make (classes());
	for (int c=0; c<result.size; c++)
		result[c] = 0;
	for (int p=0; p<patterns; p++)
		result[getClass (p)]++;
}

void PatternSource::check () const {
	ASSERT (inputs>=0);
	ASSERT (inputs<1000);
	ASSERT (outputs>=0);
	ASSERT (outputs<1000);
	ASSERT (patterns>=0);
	ASSERT (patterns<100000);
}



///////////////////////////////////////////////////////////////////////////////
//            ----                                 ----                      //
//            |   )  ___   |   |   ___        _   (      ___   |             //
//            |---   ___| -+- -+- /   ) |/\ |/ \   ---  /   ) -+-            //
//            |     (   |  |   |  |---  |   |   |     ) |---   |             //
//            |      \__|   \   \  \__  |   |   | ___/   \__    \            //
///////////////////////////////////////////////////////////////////////////////

PatternSet::PatternSet (int patts, int ins, int outs) {
	patterns = patts;
	inputs = ins;
	outputs = outs;
	// If we got the size parameters, build the structure here
	if (patts && ins && outs)
		make (patts, ins, outs);
}

PatternSet::PatternSet (const String& filename, int ins, int outs) {
	patterns=0;
	inputs=ins;
	outputs=outs;
	load (filename);
}

PatternSet::PatternSet (const Matrix& m, int mins, int mouts) {
	ASSERT (mins+mouts == m.cols || mins==-1);
	if (mins==-1)
		mins = m.cols;
	make (m.rows, mins, mouts);
	for (int r=0; r<m.rows; r++) {
		for (int c=0; c<mins; c++) 
			set_input (r,c,m.get(r,c));
		for (int c=0; c<mouts; c++)
			set_output (r,c+mins,m.get(r,c+mins));
	}
}

PatternSet::PatternSet (const PatternSet& orig) : PatternSource (orig), mInps(orig.mInps), mOutps(orig.mOutps) {
	check ();
}

void PatternSet::make (int patts, int ins, int outs) {
	// ASSERTWITH (patts>0 && ins>0 && outs>0, "Zero argument not allowed");
	PatternSource::make1 (patts, ins, outs);
	mInps.make (patts, ins);
	mOutps.make (patts, outs);
}

void PatternSet::load (const char* filename) {
	DataFormatLib::load (filename, *this);
}

void PatternSet::load (istream& in, const String& extension) {
	DataFormatLib::load (in, *this, extension);
}

void PatternSet::save (const String& filename, int filetype=FT_SNNS) {
	DataFormatLib::save (filename, *this);
}

void PatternSet::print (FILE* out) const {
	if (!out)
		out=stdout;

	for (int p=0; p<mInps.rows; p++) {
		fprintf (out, "# Input pattern %d:\n", p);
		for (int i=0; i<mInps.cols; i++)
			fprintf (out, "%f ", mInps.get (p,i));
		fprintf (out, "\n");
		fprintf (out, "# Output pattern %d:\n", p);
		for (int i=0; i<mOutps.cols; i++)
			fprintf (out, "%f ", mOutps.get (p,i));
		fprintf (out, "\n");
	}
}

void PatternSet::mutate (int errcnt) {
	int mutated[errcnt];
	int r,
		clear;

	for (int p=0; p<mInps.rows; p++)
		for (int i=0; i<errcnt; i++) {
			clear = 0;
			while (!clear) {
				r = rnd (mInps.cols)+1;
				clear=1;
				for (int j=0; j<i-1; j++)
					if (mutated[j]==r)
						clear=0;
			}
			mutated[i]		= r;
			mInps.get (p,r)	= 1-mInps.get (p,r);
		}
}

void PatternSet::recombine (int startp, int endp) {
	if (startp==-1)
		startp=0;
	if (endp==-1)
		endp=patterns-1;
	
	ASSERTWITH (startp<=endp, "Range error in recombination");

	// For each pattern
	for (int p=0; p<patterns; p++) {

		// Swap with another, random pattern

		int o=rnd (patterns);

		// Swap inputs
		for (int i=0; i<inputs; i++)
			swap (mInps.get (p, i), mInps.get (o, i));

		// Swap outputs
		for (int i=0; i<outputs; i++)
			swap (mOutps.get (p, i), mOutps.get (o, i));
	}
}

void PatternSet::recombine2 (int startp, int endp) {
	if (startp==-1)
		startp=0;
	if (endp==-1)
		endp=patterns-1;
	
	ASSERTWITH (startp<=endp, "Range error in recombination");

	// For each pattern
	for (int p=0; p<patterns; p++) {

		// Swap with another, random pattern

		int odd = p%2;
		int o=0;
		do {
			o=rnd (patterns/2)*2+odd;
		} while (o>=patterns);

		// Swap inputs
		for (int i=0; i<inputs; i++)
			swap (mInps.get (p, i), mInps.get (o, i));

		// Swap outputs
		for (int i=0; i<outputs; i++)
			swap (mOutps.get (p, i), mOutps.get (o, i));
	}
}

void PatternSet::check () const {
	PatternSource::check ();
	
}

void PatternSet::copy (const PatternSet& orig, int start, int end) {
	PatternSource::copy (orig, start, end);
}

Matrix PatternSet::getMatrix () const return result {
	result.make (patterns, inputs+outputs);
	for (int p=0; p<result.rows; p++)
		for (int c=0; c<result.cols; c++)
			if (c<inputs)
				result.get(p,c) = input(p,c);
			else
				result.get(p,c) = output(p,c-inputs);
}


///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//       _                       -----           o        ----               //
//      / \           ___          |        ___      _   (      ___   |      //
//     /   \ |/\ |/\  ___| \   |   |   |/\  ___| | |/ \   ---  /   ) -+-     //
//     |---| |   |   (   |  \  |   |   |   (   | | |   |     ) |---   |      //
//     |   | |   |    \__|   \_/   |   |    \__| | |   | ___/   \__    \     //
//                          \_/                                              //
///////////////////////////////////////////////////////////////////////////////

ArrayTrainSet::ArrayTrainSet (int ins, int outs, char* filename) {
	PatternSource::make1 (0, ins, outs);

	FILE* fin = fopen (filename, "r");
	if (!fin)
		throw new runtime_error ("file not found");

	while (!feof (fin)) {
		double	x;
		fscanf (fin, "%f", &x);
		data.add (new double (x));
	}
	fclose (fin);

	patterns = data.size-inputs-outputs;
}

void ArrayTrainSet::print (FILE* out) const {
/*	out	<< data.upper << " data items\n"
		<< patterns	<< " patterns\n"
		<< inputs	<< " inputs\n"
		<< outputs	<< " outputs\n";
	out << "t = {";
*/
	for (int i=0; i<data.size; i++) {
		fprintf (out, "%f ", data[i]);
	}
	fprintf (out, "\n");
}

void ArrayTrainSet::multiply (double k) {
	for (int i=0; i<data.size; i++)
		data[i] *= k;
}

void ArrayTrainSet::truncate (int newsize) {
	PRE (newsize<data.size && newsize>=(inputs+outputs));
	data.resize (newsize);
	patterns = data.size - inputs - outputs;
}
