/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/



#ifndef __TRAINSET_H__
#define __TRAINSET_H__

#include <magic/object.h>
#include <magic/Math.h>
#include <magic/Matrix.h>
#include <magic/attribute.h>
#include <magic/pararr.h>




//////////////////////////////////////////////////////////////////////////////
//   ----                                 ----                              //
//   |   )  ___   |   |   ___        _   (                     ___   ___    //
//   |---   ___| -+- -+- /   ) |/\ |/ \   ---   __  |   | |/\ |   \ /   )   //
//   |     (   |  |   |  |---  |   |   |     ) /  \ |   | |   |     |---    //
//   |      \__|   \   \  \__  |   |   | ___/  \__/  \__! |    \__/  \__    //
//////////////////////////////////////////////////////////////////////////////

/** The abstract base class for pattern sources such are PatternSet.
 **/
class PatternSource : public Object, public Attributed {
	decl_dynamic (PatternSource);
  public:
					PatternSource		();

	/** Inheritor's copy constructor should call this. */
					PatternSource		(const PatternSource& orig);

	/** Prints the contents of the set to the given stream. */
	virtual	void	print			(FILE* out = stdout) const {MUST_OVERLOAD}

	/** Returns the value of input variable i in pattern p. */
	virtual	double	input			(int p, int i) const {MUST_OVERLOAD; return 0.0;}

	/** Returns the value of output variable j in pattern p. */
	virtual	double	output			(int p, int j) const {MUST_OVERLOAD; return 0.0;}
	virtual void	set_input		(int p, int i, double value) {MUST_OVERLOAD;}
	virtual void	set_output		(int p, int j, double value) {MUST_OVERLOAD;}

	/** Returns the output class index for pattern p (the index of the
	 *  highest output).
	 **/
	virtual int		getClass		(int p) const;

	/** Shuffles the patterns in the set.
	 *
	 *  @param startp OPTIONAL Lower index of shuffling range.
	 *  @param endp OPTIONAL Upper index of shuffling range.
	 **/
	virtual void	recombine		(int startp=-1, int endp=-1) {MUST_OVERLOAD;}

	/** As @ref recombine above, except that this shuffles so that the
	 *  even and odd-numbered patterns are not mixed, but preserve
	 *  their evenness and oddity.
	 *
	 *  Pattern set or the shuffling range (see the parameters below)
	 *  must have even number of patterns.
	 *
	 *  @param startp OPTIONAL Lower index of shuffling range.
	 *  @param endp OPTIONAL Upper index of shuffling range.
	 **/
	virtual void	recombine2		(int startp=-1, int endp=-1) {MUST_OVERLOAD;}
	
	// Common operations
	
	/** Copies a range from another training set.
	 *
	 *  @param startp OPTIONAL Lower index of copying range.
	 *  @param endp OPTIONAL Upper index of copying range.
	 **/
	virtual void	copy			(const PatternSource& other, int startp=-1, int endp=-1);

	/** Sugar for the @ref copy operation. */
	void			operator=		(const PatternSource& other) {copy (other);}

	/** Returns the name of the set. */
	const String&	name			() const {return mName;}

	/** Sets the name of the set. */
	void			setName			(const String& name) {mName=name;}

	/** Creates this set as union of two sets.
	 **/
	void			join			(const PatternSource& a, const PatternSource& b);

	/** Splits this set into two subsets a and b according to given
	 *  ratio.
	 **/
	void			split			(PatternSource& a, PatternSource& b, double ratio) const;

	/** Copies another set by filtering the input variables with bit
	 *  string mask containing characters "0" and "1". If bits string
	 *  is null (len==0), all inputs will be taken.
	 **/
	void			filter			(const PatternSource& source, const String& bits);

	/** Returns the number of output classes in the set.
	 *
	 *  This assumes that the pattern set is for a classification
	 *  task. For each class there is one output. But, if there are
	 *  only two classes, they can be represented by just one output.
	 **/
	int				classes			() const {return (outputs==1)? 2:outputs;}

	/** Returns the number of instances for each class.
	 *
	 *  This works only for patterns sets used in classification tasks
	 *  where there is one output for each class (or just one output
	 *  in case of two classes). It is assumed that when a pattern
	 *  belongs to a certain class with index c, the c:th input
	 *  variable is set to value >0.5. Otherwise it is set to value
	 *  <0.5.
	 **/
	Array<int>		countClasses	() const;

	/** Implementation for @ref Object. */
	void			check			() const;

	/** Number of input variables in the patterns. */
	int	inputs;

	/** Number of output variables in the patterns. */
	int	outputs;

	/** Number of patterns in the pattern set. */
	int	patterns;

	/** Pattern set flags. */
	enum psetflags {INPUTS=1, OUTPUTS=2};

  protected:
	/** A name for the pattern set. */
	String	mName;

	/** Initializes a training set with given number of input and output
	 *  values and patterns.
	 **/
	virtual void	make		(int patterns, int inputs, int outputs) {MUST_OVERLOAD}
	void			make1		(int patterns, int inputs, int outputs);
	
};

#define TrainingSet PatternSource



///////////////////////////////////////////////////////////////////////////////
//            ----                                 ----                      //
//            |   )  ___   |   |   ___        _   (      ___   |             //
//            |---   ___| -+- -+- /   ) |/\ |/ \   ---  /   ) -+-            //
//            |     (   |  |   |  |---  |   |   |     ) |---   |             //
//            |      \__|   \   \  \__  |   |   | ___/   \__    \            //
///////////////////////////////////////////////////////////////////////////////

/** Typical, tabular pattern set.
 **/
class PatternSet : public PatternSource {
  public:

						PatternSet	() {}

	/** Constructs an empty pattern  set of the given dimensions.
	 **/
						PatternSet	(int patts, int ins, int outs);
	
	/** Constructs and loads a pattern set from a file. The file must
	 *  be in SNNS pattern file format. If the dimensions of the
	 *  training set have not been set, load tries to deduce them from
	 *  the file. If the dimensions of the file differ from the values
	 *  that may have been given in the construction, an exception is
	 *  thrown.
	 **/
						PatternSet	(const String& fname, int ins=0, int outs=0);

	/** Copy constructor. */
						PatternSet	(const Matrix& m, int ins=-1, int outs=0);
	
	/** Copy constructor. */
						PatternSet	(const PatternSet& orig);
				
						~PatternSet	() {}

	/** Constructs an empty pattern set with the given dimensions.
	 *
	 *  @param patts Number of patterns in the set
	 *  @param ins Number of input variables
	 *  @param outs Number of output variables
	 **/
	void				make			(int patts=0, int ins=0, int outs=0);

	enum filetypes {FT_SNNS=0, FT_RAW=1};
	
  	/** Loads a pattern set from a pattern file. Supports SNNS file
	 *  format (.pat), Proben1 (.dt), and matrix file format (other
	 *  suffixes). Overloadable to allow implementations of other
	 *  formats.
	 **/
	virtual void		load		(const char* filename);

  	/** Loads a pattern set from a pattern input stream.
	 *
	 *  @param extension File name extension; ".raw",".pat" (SNNS), or ".dt" (Proben1)
	 **/
	virtual void		load		(istream& in, const String& extension=".raw");
	
	/** Saves the training set.
	 *
	 *  @param filename Filename.
	 *  @param filetype One of FT_SNNS (SNNS format, default), FT_RAW (raw format).
	 **/
	void				save			(const String& filename, int filetype=FT_SNNS);
	
	// Virtual method implementations

	virtual void		print			(FILE* out = stdout) const;
	virtual double		input			(int p, int i) const {return mInps.get (p,i);}
	virtual double		output			(int p, int j) const {return mOutps.get (p,j);}
	virtual void		set_input		(int p, int i, double value) {mInps.get (p,i) = value;}
	virtual void		set_output		(int p, int j, double value) {mOutps.get (p,j) = value;}

	Matrix				getMatrix		() const;

	/** Mutates the inputs of the training set by given number of
	 *  errors. The number of errors equals to the hamming distance
	 *  between normal and mutated patterns.
	 **/
	void				mutate			(int errcnt);

	/** Permutates a set of patterns to a random configuration. If range
	 *  is specified, only those patterns are reshuffled.
	 **/
	void				recombine		(int startp=-1, int endp=-1);

	/** As above, except ... */
	void				recombine2		(int startp=-1, int endp=-1);
	
	/** Implementation for @ref PatternSource. Copies a range from
	 *  another training set.
	 *
	 *  @param startp OPTIONAL Lower index of copying range.
	 *  @param endp OPTIONAL Upper index of copying range.
	 **/
	virtual void		copy			(const PatternSet& orig, int start=-1, int end=-1);
	
	void				check			() const;

  protected:
	/** Input patterns */
	Matrix			mInps;

	/** Output patterns */
	Matrix			mOutps;

	/** Optional comments in patterns. */
	//Array<String>*	mComments;

	/** Calculates the means of each input and output field. */
	void			calc_means		();

	/** Calculates the standard deviations of each input and output field. */
	void			calc_stddevs	();

  private:
	operator= (const PatternSet& orig) {FORBIDDEN}
};



///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//       _                       -----           o        ----               //
//      / \           ___          |        ___      _   (      ___   |      //
//     /   \ |/\ |/\  ___| \   |   |   |/\  ___| | |/ \   ---  /   ) -+-     //
//     |---| |   |   (   |  \  |   |   |   (   | | |   |     ) |---   |      //
//     |   | |   |    \__|   \_/   |   |    \__| | |   | ___/   \__    \     //
//                          \_/                                              //
///////////////////////////////////////////////////////////////////////////////

/** Data series training set, especially useful for time series. This
 *  class was made to make it possible to use plain data sets as
 *  training sets without the need to first convert their source files
 *  to classical training set files. The class simply reads the set
 *  from file to an array, and then calculates the values for input
 *  and output requests relatively. It also calculates the number of
 *  training patterns from the size of the set and sizes of inputs and
 *  outputs.
 **/
class ArrayTrainSet : public PatternSource {
  public:

	/** Standard constructor.
	 *
	 *  @param inputs Number of input variables.
	 *  @param outputs Number of output variables.
	 *  @param filename Filename of the file that contains the series.
	 **/
					ArrayTrainSet	(int inputs, int outputs, char* filename);
					~ArrayTrainSet	() {;}

	/** FORBIDDEN */
	virtual void	make			(int patterns, int inputs, int outputs) {FORBIDDEN}

	// Virtual method implementations
	
	virtual void	print			(FILE* out = stdout) const;
	virtual double	input			(int p, int i) const {return data[p+i];}
	virtual double	output			(int p, int j) const {return data[p+inputs+j];}

	// New methods
	
	/** Scales all input and output values with the multiplier k.
	 **/
	void			multiply		(double k);

	/** Resizes the dataset to the given size.
	 **/
	void			truncate		(int size);

  public:
	/** Here all the patterns are represented in the same array. First
	 *  pattern is at data[0..ins-1], second at data[1..ins-0], third
	 *  at data[2..ins+1], etc.
	 **/
	Array<double>	data;

};

#endif
