/***************************************************************************
 *   copyright            : (C) 2000 by Marko Grnroos                     *
 *   email                : magi@iki.fi                                    *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#ifndef __TOPOLOGY_H__
#define __TOPOLOGY_H__

#include "annetwork.h"


///////////////////////////////////////////////////////////////////////////////
//           _   |   | |   | -----                |                          //
//          / \  |\  | |\  |   |         --       |                          //
//         /   \ | \ | | \ |   |    __  |  )  __  |  __   ___  \   |         //
//         |---| |  \| |  \|   |   /  \ |--  /  \ | /  \ (   \  \  |         //
//         |   | |   | |   |   |   \__/ |    \__/ | \__/  ---/   \_/         //
//                                                        __/   \_/          //
///////////////////////////////////////////////////////////////////////////////

/** Baseclass for builder methods for ANNetworks.
 *
 *  Patterns: Builder (builds ANNs), Strategy.
 **/
class ANNTopology {
  public:
	virtual			~ANNTopology ();
	
	virtual void	build	(ANNetwork& network) const=0;
	virtual void	empty	()=0;
  protected:
};


///////////////////////////////////////////////////////////////////////////////
//          _   |   | |   | |                           o                    //
//         / \  |\  | |\  | |      ___         ___          _                //
//        /   \ | \ | | \ | |      ___| \   | /   ) |/\ | |/ \   ___         //
//        |---| |  \| |  \| |     (   |  \  | |---  |   | |   | (   \        //
//        |   | |   | |   | |____  \__|   \_/  \__  |   | |   |  ---/        //
//                                       \_/                     __/         //
///////////////////////////////////////////////////////////////////////////////

/** Older, layered ANN topology. Currently semi-obsolete, but still
 *  used by various parts of Inanna.
 *
 **/
class ANNLayering : public Object {
  public:

	/** Standard constructor.
	 *
	 *  See @ref make for info about the parameters.
	 **/
				ANNLayering		(const char* description=NULL) {make (description);}
	virtual		~ANNLayering	();

	/** Sets the layer structure according to the given description string.
	 *
	 *  @param description Textual layer description. Consists of
	 *  layer sizes separated by '-'-marks. For example: "2-10-10-1".
	 *  NOTE: The layer sizes may currently NOT have additional
	 *  qualifiers that are not numbers, for example:
	 *  "S2-10-10-L10". These qualifiers are usually intended for the
	 *  @ref AnyNetwork inheritor, and should ignored here, but, as
	 *  said, are currently not.
	 **/
	void		make			(const char* description);

	/** Calculates the absolute neuron index offset for the given
	 *  layer index; index of the first neuron on that layer. If the
	 *  given layer index is negative, it is substracted from the
	 *  number of layers (-1 means the output layer, -2 the layer
	 *  below that, etc).
	 **/
	int			layerIndex		(int layer) const;

	/** Calculates the relative layer and offset for given absolute
	 *  unit index.
	 *
	 *  @param i Neuron index.
	 *
	 *  @param layerno Reference-returned result: layer index where
	 *  the neuron resides.
	 *
	 *  @param layerpos Reference-returned result: position of the
	 *  neuron on the layer indicated by the layerno result parameter.
	 **/
	void		getPos			(int i, int& layerno, int& layerpos) const;

	/** Returns the total number of units in the network.
	 **/
	int			totalUnits		() const {
		return layerIndex(mLayers.size-1)+mLayers[mLayers.size-1];}

	/** Returns the number of layers in the network.
	 **/
	int			layers			() const {return mLayers.size;}

	/** Converts the layering back to layering description string.
	 **/
	String		toString		() const;

	/** Returns the size of the layer i. Const version.
	 *
	 * If the given layer index is negative, it is substracted from
	 * the number of layers (-1 means the output layer, -2 the layer
	 * below that, etc).
	 **/
	const int&	operator[]		(int i) const {return mLayers [(i<0)?mLayers.size+i:i];}

	/** Returns the size of the layer i. Non-const version.
	 *
	 * If the given layer index is negative, it is substracted from
	 * the number of layers (-1 means the output layer, -2 the layer
	 * below that, etc).
	 **/
	int&		operator[]		(int i) {return mLayers [(i<0)?mLayers.size+i:i];}

	/** Clears the layering; makes the network description empty.
	 **/
	void		empty			() {mLayers.empty();}

	/** Standard copy operator. */
	void		operator=		(const ANNLayering& o) {mLayers=o.mLayers;}

	/** Removes a layer. NOTE: The negative indexing is not allowed
	 *  here.
	 **/
	void		removeLayer		(int i) {mLayers.removeFill(i);}

  protected:
	Array<int>	mLayers;
};



/////////////////////////////////////////////////////////////////////////////////////
// |                                     | -----                |                  //
// |      ___         ___       ___      |   |         --       |                  //
// |      ___| \   | /   ) |/\ /   )  ---|   |    __  |  )  __  |  __   ___  \   | //
// |     (   |  \  | |---  |   |---  (   |   |   /  \ |--  /  \ | /  \ (   \  \  | //
// |____  \__|   \_/  \__  |    \__   ---|   |   \__/ |    \__/ | \__/  ---/   \_/ //
//              \_/                                                     __/   \_/  //
/////////////////////////////////////////////////////////////////////////////////////

/** Multilayer perceptron ANNTopology.
 **/
class LayeredTopology : public ANNTopology, public ANNLayering {
  public:

	/** Standard constructor.
	 *
	 *  See @ref make for info about the parameters.
	 **/
					LayeredTopology		(const char* description=NULL) {make (description);}
	virtual			~LayeredTopology	();

	virtual void	build	(ANNetwork& network) const;

	/** Clears the layering; makes the network description empty.
	 **/
	virtual void	empty			() {ANNLayering::empty();}

  protected:
};


#endif
