/***************************************************************************
 *   copyright            : (C) 2000 by Marko Grnroos                     *
 *   email                : magi@iki.fi                                    *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#include "trainer.h"
#include "termination.h"
#include "patternset.h"

///////////////////////////////////////////////////////////////////////////////
//                     -----           o                                     //
//                       |        ___      _    ___                          //
//                       |   |/\  ___| | |/ \  /   ) |/\                     //
//                       |   |   (   | | |   | |---  |                       //
//                       |   |    \__| | |   |  \__  |                       //
///////////////////////////////////////////////////////////////////////////////

Trainer::Trainer () {
	mTerminatorName = "GL2";
	mGeneralizationLoss = 0.0;
	mTrained=0;
	mTotalTrained=0;
	pTrainingObserver=NULL;
}

/*virtual*/ void Trainer::init (const StringMap& params) {
	INITPARAMS(params, 
			   );
}

/*virtual*/ void Trainer::initTrain (ANNetwork& network) const {
	// Initialize weights
	network.init (0.5);
}

double Trainer::train (ANNetwork& network, const PatternSet& trainset, int cycles, const PatternSet* validationSet, int validationInterval) {
	ASSERT (trainset.patterns>0);
	ASSERT (cycles>0);

	// Initialize training method
	initTrain (network);

	// Initialize recording
	mTrainingProfile.make (cycles);
	for (int i=0; i<mTrainingProfile.size; i++)
		mTrainingProfile[i]=0.0;

	Terminator* arnold=NULL;
	bool ensureValidGTTrain = true;
	if (validationSet) {
		mValidationProfile.make (cycles/validationInterval+1);
		for (int i=0; i<mValidationProfile.size; i++)
			mValidationProfile[i]=0.0;

		////////////////////////////////////////
		// Determine termination method
		
		String terminator = mTerminatorName;
		
		// Check if we should _not_ check for that validError>trainError
		if (terminator.left(1) == "-") {
			ensureValidGTTrain = false;
			terminator = terminator.substr (1);
		}
		
		// Can't use termination if there are no validation patterns...
		if (validationSet->patterns==0)
			terminator = "none";
		
		// Order a terminator from the factory
		arnold = buildTerminator (terminator, *validationSet, validationInterval);
	}

	////////////////////////////////////////
	// Train and validate
	
	double	trainMSE;
	double	GL			= 0;
	int		validations	= 0;
	bool	terminate	= false;
	for (mTotalTrained=0; mTotalTrained<cycles;) {
		// Train all patterns once
		trainMSE = mTrainingProfile[mTotalTrained] = trainOnce (network, trainset);
		mTotalTrained++;

		// Streamed output
		if (false)
			printf ("Cycle %d MSE=%f\n", mTotalTrained, trainMSE);
		
		// Validate for early stopping
		if (arnold && mTotalTrained>0 && !(mTotalTrained%validationInterval)) {

			// Calculate the validation error for the current network
			// state
			terminate = arnold->validate (network, *this, mTotalTrained);
			mValidationProfile[validations++] = arnold->validationError ();

			// Let the terminator calculate the GL value. Some
			// terminators use this value to determine termination.
			GL = arnold->generalizationLoss ();

			// Do not terminate if the validation error is lower than
			// the training error
			if (ensureValidGTTrain && arnold->validationError() < trainMSE)
				break;
		}

		if (pTrainingObserver) {
			pTrainingObserver->cycleTrained (*this, mTotalTrained);
			if (pTrainingObserver->wantsToStop())
				break;
		}

	}
	//printf ("tp.size=%d, mTrained=%d, mTotalTrained=%d, arnold=%d, arnolderr=%f\n",
	//		mTrainingProfile.size, mTrained, mTotalTrained, arnold->howManyTrained(),
	//		arnold->minimumError());

	// Restore the state with the lowest error on validation set. Do
	// not restore if the validation error is smaller than the training error
	if (arnold && (!ensureValidGTTrain || arnold->minimumError() > trainMSE)) {
		arnold->restore (network);
		mTrained = arnold->howManyTrained ();
	} else
		// No validation, keep the latest state
		mTrained = mTotalTrained;
		
	// Record and clean up some statistics (truncate vectors)
	mGeneralizationLoss = GL;
	if (mTrained>0)
		mTrainingProfile.resize (mTrained);
	else
		mTrainingProfile.make (0);
	if (arnold)
		mValidationProfile.resize (validations);
	else
		mValidationProfile.make (0);

    // You will now be terminated
    delete arnold;

	return trainMSE; // Return final training MSE
}



////////////////////////////////////////////////////////////////////////////////
// ----              |                      -----           o                 //
// |   )  ___   ___  |    --            --    |        ___      _    ___      //
// |---   ___| |   \ | / |  ) |/\  __  |  )   |   |/\  ___| | |/ \  /   ) |/\ //
// |   ) (   | |     |/  |--  |   /  \ |--    |   |   (   | | |   | |---  |   //
// |___   \__|  \__/ | \ |    |   \__/ |      |   |    \__| | |   |  \__  |   //
////////////////////////////////////////////////////////////////////////////////

/*virtual*/ void BackpropTrainer::init (const StringMap& params) {
	Trainer::init (params);
	INITPARAMS(params, 
		mEta			= params["BackpropTrainer.eta"];
		mMomentum		= params["BackpropTrainer.momentum"];
		mDecay			= params["BackpropTrainer.decay"];
		mBatchLearning	= int(params["BackpropTrainer.batchLearning"]);
		);
}

/*virtual*/ Array<DynParameter>* BackpropTrainer::parameters () const return result {
	result = new Array<DynParameter>;
	result->add (new DoubleParameter	("eta", "Learning rate", 15, 0.0, 1.0, 0.25));
	result->add (new DoubleParameter	("momentum", "Weight momentum", 15, 0.0, 1.0, 0.9));
	result->add (new DoubleParameter	("decay", "Weight decay multiplier", 15, 0.5, 1.0, 1.0));
	result->add (new BoolParameter		("batchLearning", "Update weights in batch"));
}

/*virtual*/ void BackpropTrainer::initTrain (ANNetwork& network) const {
	Trainer::initTrain (network);
	
	// Count the total number of connections in the entire network
	int connections=0;
	for (int i=0; i<network.size; i++)
		connections += network[i].incomings();

	// Create weight delta data for connections and biases
	mWeightDeltas.make (connections + network.size);
	for (int i=0; i<mWeightDeltas.size; i++)
		mWeightDeltas[i] = 0.0;
}

/*virtual*/ double BackpropTrainer::trainOnce (ANNetwork& network, const PatternSet& set) const {
	/*
	if (mBatchLearning)
		for (int i=0; i<mWeightDeltas.size; i++)
			mWeightDeltas[i] = 0.0;
	*/
	
	// Train each pattern once
	double sse=0.0;
	for (int p=0; p<set.patterns; p++)
		sse += trainPattern (network, set, p);

	if (true || mBatchLearning)
		updateWeights (network);

	// Actualize weight adjustments
	/*
	if (mBatchLearning)
		for (int n=0, oc=0; n<network.size; n++)
			for (int c=0; c<network[n].incomings(); c++, oc++)
				network[n].incoming(c).setWeight (mWeightDeltas[oc]);
	*/
	return sse/set.patterns; // Return MSE
}

/*virtual*/ double BackpropTrainer::trainPattern (ANNetwork& network, const PatternSet& set, int p) const {
	// Feed the pattern to the network
	for (int inp=0; inp<set.inputs; inp++)
		network[inp].setActivation (set.input (p, inp));

	// Forward pass
	network.update ();

	// Backward pass
	backpropagate (network, set, p);

	// Calculate error
	double sse=0.0;
	for (int outp=0; outp<set.outputs; outp++)
		sse += sqr(set.output(p,outp) - network[network.size-set.outputs+outp].activation());

	return sse / set.outputs; // Return MSE
}

void BackpropTrainer::backpropagate (register ANNetwork& network, register const PatternSet& set, int p) const {
	mError.make (network.size);
	int outLayerBase = network.size-set.outputs;
	register Connection* conn;
	register double sum_k;
	register Neuron* neuron_j;
	register double delta_j;
	register int j, k;
	
	// Iterate backwards
	for (j=network.size-1; j>=0; j--) {
		neuron_j = &network[j];
		// Calculate error at a neuron
		if (j>=outLayerBase) { // Output neuron
			delta_j = (set.output(p,j-outLayerBase) - neuron_j->activation())
				* neuron_j->activation() * (1.0 - neuron_j->activation());
		} else { // A hidden or input neuron
			sum_k=0.0;
			for (int k=0; k<neuron_j->outgoings(); k++)
				sum_k += mError [neuron_j->outgoing(k).target().id()] * neuron_j->outgoing(k).weight();
			delta_j = neuron_j->activation()*(1.0-neuron_j->activation()) * sum_k;
		}
		mError[j] = delta_j;
	}
}

void BackpropTrainer::updateWeights (register ANNetwork& network) const {
	register int j, i, ji;
	register double deltaw_ji;
	
	for (j=network.size-1, ji=0; j>=0; j--) {

		// Update bias
		deltaw_ji = mEta * mError[j];
		network[j].setBias (network[j].bias() + deltaw_ji + mMomentum*mWeightDeltas[ji]);
		mWeightDeltas[ji++] = deltaw_ji;

		// Update weights
		for (i=network[j].incomings()-1; i>=0; i--, ji++) {
			register Connection& conn = network[j].incoming(i);
			deltaw_ji = mEta * mError[j] * conn.source().activation();
			conn.setWeight (conn.weight() + deltaw_ji + mMomentum*mWeightDeltas[ji]);
			mWeightDeltas[ji] = deltaw_ji;
		}
	}
}



///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//        ----  ----                -----           o                        //
//        |   ) |   )           --    |        ___      _    ___             //
//        |---  |---  |/\  __  |  )   |   |/\  ___| | |/ \  /   ) |/\        //
//        | \   |     |   /  \ |--    |   |   (   | | |   | |---  |          //
//        |  \  |     |   \__/ |      |   |    \__| | |   |  \__  |          //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

/*virtual*/ void RPropTrainer::init (const StringMap& params) {
	Trainer::init (params);
	
	INITPARAMS(params, 
		mDelta0			= params["RPropTrainer.delta0"];
		mDeltaMax		= params["RPropTrainer.deltamax"];
		mDecay			= params["BackpropTrainer.decay"];
		mBatchLearning	= int(params["BackpropTrainer.batchLearning"]);
		);
}

/*virtual*/ Array<DynParameter>* RPropTrainer::parameters () const return result {
	result = new Array<DynParameter>;
	result->add (new DoubleParameter	("delta0", "Initial learning rate", 15, 0.0, 100.0, 0.1));
	result->add (new DoubleParameter	("deltamax", "Maximum learning rate", 15, 0.0, 100.0, 50.0));
	result->add (new DoubleParameter	("decay", "Weight decay multiplier", 15, 0.5, 1.0, 1.0));
	result->add (new IntParameter		("maxCycles", "Max training cycles", 1, 100000, 100));
	result->add (new BoolParameter		("batchLearning", "Update weights in batch"));
}

/*virtual*/ void RPropTrainer::initTrain (ANNetwork& network) const {
	BackpropTrainer::initTrain (network);
	
	// Set initial deltas
	mDelta.make (mWeightDeltas.size);
	for (int i=0; i<mDelta.size; i++)
		mDelta[i] = mDelta0;

	// Initialize "previous neuron error"
	mGradient.make (mWeightDeltas.size);
	for (int i=0; i<mGradient.size; i++)
		mGradient[i] = 0.0;
}

inline double sign (double x) {return (x>=0)? 1:-1;}
inline double min (double x, double y) {return (x>y)? y:x;}

// This algorithm is largely copied from the SNNS implementation,
// found in SNNS/kernel/sources/learn_f.c
//
// SNNS variable key:
//  value_a = mDelta[ji] = "update value" = error_ji = 
//  value_b = mOldDeltaW[ji] = delta_w(t-1)
//  value_c = gradient_ji = sum(dEdw)

/*virtual*/ void RPropTrainer::backpropagate (register ANNetwork& network, register const PatternSet& set, int p) const {
	BackpropTrainer::backpropagate (network, set, p);
	
	// Calculate per-weight errors
	for (register int j=network.size-1, ji=0; j>=0; j--)
		for (register int i=-1; i<network[j].incomings(); i++, ji++)
			if (i==-1) // Bias
				mGradient[ji] -= mError[j];
			else // Weight
				mGradient[ji] -= mError[j] * network[j].incoming(i).source().activation();
}

Connection nullconn;

/*virtual*/ void RPropTrainer::updateWeights (ANNetwork& network) const {
	int j;
	int ji;
	int i;
	double gradient_ji;
	double direction;
	
	for (j=network.size-1, ji=0; j>=0; j--) {
		// Update weights
		for (i=-1; i<network[j].incomings(); i++, ji++) {
			Connection& conn = (i==-1)? network[j].getBiasObj() : network[j].incoming(i);
			double& delta = mDelta[ji];

			// Weight decay
			gradient_ji = mGradient[ji] + (1-mDecay)*conn.weight();

			// Calculate dw * dEdw
			direction = gradient_ji * mWeightDeltas[ji];

			if (direction < 0.0) {			// Same direction: dw * dEdw < 0
				delta *= 1.2;
				if (delta > mDeltaMax)
					delta = mDeltaMax;
				if (gradient_ji < 0.0)
					mWeightDeltas[ji] =  delta;
				else
					mWeightDeltas[ji] = -delta;
			} else if (direction > 0.0) {	// Direction changed
				mWeightDeltas[ji] = 0.0;
				delta *= 0.5;
				if (delta < 1E-6)
					delta = 1E-6;
			} else {						// Start of RProp learning process
				if (gradient_ji<0.0)
					mWeightDeltas[ji] = delta;
				else
					mWeightDeltas[ji] = -delta;
			}

			// Update weight or bias
			if (i==-1)
				network[j].setBias (network[j].bias() + mWeightDeltas[ji]);
			else
				conn.setWeight (conn.weight() + mWeightDeltas[ji]);
			mGradient[ji] = 0.0;
		}
	}
}

// 
