/***************************************************************************
 *   copyright            : (C) 2000 by Marko Grnroos                     *
 *   email                : magi@iki.fi                                    *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#ifndef __TRAINER_H__
#define __TRAINER_H__

#include <magic/parameter.h>
#include "freenet.h"

// Locals:
class TrainingObserver;

///////////////////////////////////////////////////////////////////////////////
//                     -----           o                                     //
//                       |        ___      _    ___                          //
//                       |   |/\  ___| | |/ \  /   ) |/\                     //
//                       |   |   (   | | |   | |---  |                       //
//                       |   |    \__| | |   |  \__  |                       //
///////////////////////////////////////////////////////////////////////////////

/** Abstract superclass for neural network training algorithms.
 *
 *  Desing Patterns: Strategy.
 **/
class Trainer : public Object, public IParameterized {
  public:
							Trainer			();

	virtual void			init			(const StringMap& params);

	virtual double			train			(ANNetwork& network, const PatternSet& trainset, int cycles, const PatternSet* validationSet=NULL, int validationInterval=0);
	
	/** Sets the termination method by name (UP2, GL5, etc).
	 *
	 *  See @ref Terminator, it's inheritors and the global
	 *  buildTerminator() method for more info about the parameters.
	 **/
	void					setTerminator	(const String& name) {mTerminatorName=name;}

	// Informative methods
	
	/** Returns the number of training cycles the network has been
	 *  trained so far.
	 *
	 *  The network may actually have been trained more than this, but
	 *  an earlier weight state may have been restored by a @ref
	 *  Terminator (early stopping method).
	 **/
	int						cyclesTrained	() const {return mTrained;}

	/** Total number of training cycles, including any cycles cut out
	 *  by a @ref Terminator.
	 *
	 *  @see LearningMapping::cyclesTrained
	 **/
	int						totalCycles		() const {return mTotalTrained;}

	/** Returns the percentual error growth between lowest validation
	 *  error and the validation error at the end of the training.
	 **/
	double					generalizLoss	() const {return mGeneralizationLoss;}

	/** This method is used by the @ref Terminator method to store the
	 *  generalization loss value here.
	 **/
	void					setGeneralizLoss(double gl) {mGeneralizationLoss = gl;}

	/** Returns a history record of training set error during
	 *  training. Useful for both statistical analysis of learning
	 *  curves and possibly also some @ref Terminator methods.
	 **/
	const Vector&			trainingRecord	() const {return mTrainingProfile;}
	
	/** Returns a history recording of validation set error during
	 *  training. Useful for both statistical analysis of learning
	 *  curves and possibly also some @ref Terminator methods.
	 **/
	const Vector&			validationRecord() const {return mValidationProfile;}

	void					setObserver		(TrainingObserver* observer) {pTrainingObserver=observer;}
	
  protected:

	/** Initializes training. */
	virtual void			initTrain		(ANNetwork& network) const;

	/** Trains the pattern set once. */
	virtual double			trainOnce		(ANNetwork& network, const PatternSet& set) const {MUST_OVERLOAD; return 0.0;}

  protected:

	/** Name of the current termination method. */
	String	mTerminatorName;

	/** Number of relevant training cycles trained so far. The network
	 *  may have been trained more than this, but an earlier weight
	 *  state may have been restored by a @ref Terminator (early
	 *  stopping method).
	 **/
	int		mTrained;

	/** Total number of training cycles, including any cycles cut out
	 *  by a @ref Terminator.
	 *
	 *  @see LearningMapping::mTrained
	 **/
	int		mTotalTrained;

	/** Current loss of generalization ability of the network in
	 *  respect to the state with the lowest validation error.
	 *
	 *  This value changes during the training if some @ref Terminator
	 *  which uses the GL measurement is used.
	 *
	 *  Generalization loss is calculated by dividing last fitness by
	 *  the best validation error.
	 **/
	double	mGeneralizationLoss;

	/** History record of training set error during training. Useful
	 *  for both statistical analysis of learning curves and possibly
	 *  also some @ref Terminator methods.
	 **/
	Vector	mTrainingProfile;
	
	/** Recording of validation set error during training. Useful for
	 *  both statistical analysis of learning curves and possibly also
	 *  some @ref Terminator methods.
	 **/
	Vector	mValidationProfile;

	/** Observer object that gets called after every cycle.
	 */
	TrainingObserver*	pTrainingObserver;
	
	friend Terminator;
};


////////////////////////////////////////////////////////////////////////////////
// ----              |                      -----           o                 //
// |   )  ___   ___  |    --            --    |        ___      _    ___      //
// |---   ___| |   \ | / |  ) |/\  __  |  )   |   |/\  ___| | |/ \  /   ) |/\ //
// |   ) (   | |     |/  |--  |   /  \ |--    |   |   (   | | |   | |---  |   //
// |___   \__|  \__/ | \ |    |   \__/ |      |   |    \__| | |   |  \__  |   //
////////////////////////////////////////////////////////////////////////////////

/** Error back propagation neural learning algorithm.
 *
 *  Design Patterns: Template Method (various parts of the algorithm
 *  can be overloaded).
 **/
class BackpropTrainer : public Trainer {
  public:
	virtual Array<DynParameter>*	parameters	() const;
	virtual void					init		(const StringMap& params);
	
  protected:
	/** Initializes training. */
	virtual void					initTrain		(ANNetwork& network) const;

	/** Implementation for Trainer. */
	virtual double					trainOnce		(ANNetwork& network, const PatternSet& set) const;

	/** Trains one pattern. */
	virtual double					trainPattern	(ANNetwork& network, const PatternSet& set, int p) const;

	/** Propagates an error signal backwards in the network. Does not
	 *  modify the network in any way, but stores the per-neuron error
	 *  in mError.
	 **/
	virtual void					backpropagate	(ANNetwork& network, const PatternSet& set, int p) const;

	/** Updates weights after backpropagation phase. */
	virtual void					updateWeights	(ANNetwork& network) const;

  protected:
	/** Learning speed. */
	double	mEta;

	/** Momentum. */
	double	mMomentum;

	/** Weight decay multiplier. */
	double	mDecay;

	/** Should batch learning be used? */
	bool	mBatchLearning;

	mutable Vector	mWeightDeltas;
	mutable Vector	mError;
};



///////////////////////////////////////////////////////////////////////////////
//        ----  ----                -----           o                        //
//        |   ) |   )           --    |        ___      _    ___             //
//        |---  |---  |/\  __  |  )   |   |/\  ___| | |/ \  /   ) |/\        //
//        | \   |     |   /  \ |--    |   |   (   | | |   | |---  |          //
//        |  \  |     |   \__/ |      |   |    \__| | |   |  \__  |          //
///////////////////////////////////////////////////////////////////////////////

/** Resilient error backpropagation algorithm by Riedmiller.
 *
 *  Design Patterns: Template Method (various parts of the algorithm
 *  can be overloaded).
 **/
class RPropTrainer : public BackpropTrainer {
  public:
	virtual Array<DynParameter>*	parameters	() const;
	virtual void					init		(const StringMap& params);
	
  protected:

	/** Implementation for BackpropTrainer. Initializes training. */
	virtual void					initTrain		(ANNetwork& network) const;

	/** Implementation for BackpropTrainer. */
	virtual void					backpropagate	(ANNetwork& network, const PatternSet& set, int p) const;

	/** Updates weights after backpropagation phase. */
	virtual void					updateWeights	(ANNetwork& network) const;

  protected:
	/** Initial per-weight delta. */
	double	mDelta0;

	/** Maximum per-weight delta. */
	double	mDeltaMax;

	/** Per-weight deltas. */
	mutable Vector	mDelta;

	/** Per-weight ... */
	mutable Vector	mOldDeltaW;

	/** Per-weight summed error */
	mutable Vector	mGradient;
};



///////////////////////////////////////////////////////////////////////////////////////
// -----           o       o              ___                                        //
//   |        ___      _       _         |   | |      ____  ___             ___      //
//   |   |/\  ___| | |/ \  | |/ \   ___  |   | |---  (     /   ) |/\ |   | /   ) |/\ //
//   |   |   (   | | |   | | |   | (   \ |   | |   )  \__  |---  |    \ /  |---  |   //
//   |   |    \__| | |   | | |   |  ---/ `___ |__/  ____)  \__  |     V    \__  |   //
//                                  __/                                              //
///////////////////////////////////////////////////////////////////////////////////////

class TrainingObserver : public Object {
  public:
					TrainingObserver	() {mStop = false;}
	
	void			initTraining		() {mStop = false;}
	bool			wantsToStop			() const {return mStop;}
	
	virtual	void	cycleTrained		(const Trainer& trainer, int totalCycles)=0;
	
  protected:
	void			stopTraining		() {mStop=true;}
	
  private:
	bool	mStop;
};

#endif
